﻿using System.Linq;
using Infragistics.Olap.Adomd.AnalysisAccess;

namespace Infragistics.Olap.Adomd
{
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using Infragistics.Olap;
    using Infragistics.Olap.Data;
    using Infragistics.Olap.Data.Base;
    using Infragistics.Olap.Xmla;
    using Member = Infragistics.Olap.Data.Adomd.Member;

    public class AdomdCommand : XmlaCommand
    {
        public AdomdCommand(XmlaConnection connection)
            : base(connection)
        {
        }

        public override void ExecuteCommandAsync(IMdxQuery query, XmlaConnection connection)
        {
            string commandText = GenerateCommand(query);
            if (this.CancelPending)
            {
                return;
            }

            AdomdConnectorClient client = new AdomdConnectorClient();

            client.ExecuteCommandCompleted += (s, args) =>
            {
                IResult result = null;

                if (args.Error == null && !args.Cancelled)
                {
                    AxisData columnAxis = null;
                    if (query.ColumnAxis != null)
                    {
                        columnAxis = args.Result.Axes[0];
                    }

                    AxisData rowAxis = null;
                    if (query.RowAxis != null)
                    {
                        rowAxis = args.Result.Axes[columnAxis == null ? 0 : 1];
                    }

                    result = GenerateResult(columnAxis, rowAxis, args.Result.Cells, query.Cube);
                }

                this.OnExecuteCommandCompleted(new Infragistics.Olap.ExecuteCommandCompletedEventArgs(result, args.Error, args.Cancelled, query));
            };

            client.ExecuteCommandAsync(((AdomdConnectionSettings)connection.ConnectionSettings).ConnectionString, commandText);
        }

        private static Result GenerateResult(AxisData columnAxis, AxisData rowAxis, IList<CellData> cellsData, ICube cube)
        {
            ResultAxis resultColumnAxis = GenerateAxis(columnAxis, cube);

            ResultAxis resultRowAxis = GenerateAxis(rowAxis, cube);

            ICell[,] cells = new ICell[resultRowAxis.Tuples.Count, resultColumnAxis.Tuples.Count];
            FillCellData(cells, cellsData);

            if (resultColumnAxis == ResultAxis.Empty &&
                resultRowAxis == ResultAxis.Empty &&
                cells[0, 0] == null)
            {
                return Result.Empty;
            }

            return new Result(resultRowAxis, resultColumnAxis, cells);
        }

        private static void FillCellData(ICell[,] cells, IList<CellData> cellsData)
        {
            if (cellsData == null)
            {
                return;
            }

            int columnsCount = cells.GetLength(1);

            for (int i = 0; i < cellsData.Count; i++)
            {
                CellData cellData = cellsData[i];
                int columnNumber = i % columnsCount;
                int rowNumber = i / columnsCount;

                cells[rowNumber, columnNumber] = new Cell(cellData.Value, cellData.FormattedValue);
            }
        }

        private static ResultAxis GenerateAxis(AxisData axis, ICube cube)
        {
            List<ITuple> tuples = new List<ITuple>();
            int[] maxDepths;

            if (axis == null || axis.Positions.Count == 0)
            {
                return ResultAxis.Empty;
            }

            IEnumerable<ILevel> levels = cube.Dimensions.SelectMany(d => d.Hierarchies).SelectMany(h => h.Levels);
            PositionData prevTuple = null;
            foreach (PositionData positionData in axis.Positions)
            {
                List<IMember> members = new List<IMember>();
                int memberIndex = 0;

                foreach (MemberData memberData in positionData.Members)
                {
                    bool isTotal = false;
                    if (memberData.DrilledDown)
                    {
                        // member.DrillDownMember = true;
                        isTotal = true;
                    }
                    else
                    {
                        // the member from the previous tuple is child of this member
                        // that does mean it is the Total column and it follows its children members                          
                        if (!memberData.ParentSameAsPrevious &&
                            prevTuple != null &&
                            prevTuple.Members[memberIndex].ParentUniqueName == memberData.UniqueName)
                        {
                            isTotal = true;
                        }
                    }

                    string name = memberData.Name;
                    string uniqueName = memberData.UniqueName;
                    string caption = memberData.Caption;
                    string parentUniqueName = memberData.ParentUniqueName;
                    string levelName = memberData.LevelName;
                    int levelDepth = memberData.LevelDepth;
                    int childCount = memberData.ChildCount;
                    bool drilledDown = memberData.DrilledDown;
                    bool parentSameAsPrevious = memberData.ParentSameAsPrevious;

                    ILevel parentLevel = levels.FirstOrDefault(l => l.UniqueName == levelName);
                    IMember originalMember = null;
                    Member member;
                    if (parentLevel == null || !parentLevel.TryGetMember(uniqueName, out originalMember))
                    {
                        member = new Member(
                            name, uniqueName, caption, parentUniqueName, levelName, levelDepth, parentLevel, childCount,
                            drilledDown, parentSameAsPrevious, isTotal);
                    }
                    else
                    {
                        member = new Member(
                            name, uniqueName, caption, parentUniqueName, levelName, levelDepth, parentLevel, childCount,
                            drilledDown, parentSameAsPrevious, isTotal, originalMember.UniqueId);
                    }

                    members.Add(member);

                    memberIndex++;
                }

                prevTuple = positionData;

                Infragistics.Olap.Data.Xmla.SqlAnalysisServices2008.Tuple axisTuple =
                    new Infragistics.Olap.Data.Xmla.SqlAnalysisServices2008.Tuple(new OlapElementCollection<IMember>(members));

                tuples.Add(axisTuple);
            }

            return new ResultAxis(new ReadOnlyCollection<ITuple>(tuples));
        }
    }
}
